#include "ehs/io/File_W32.h"

namespace ehs
{
	File::~File()
	{
		if (view && !UnmapViewOfFile(view))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to unmap view with error #" + Str_8::FromNum(GetLastError()) + ".");

		if (map != INVALID_HANDLE_VALUE && !CloseHandle(map))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to unmap with error #" + Str_8::FromNum(GetLastError()) + ".");

		if (hdl != INVALID_HANDLE_VALUE && !CloseHandle(hdl))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to close file handle with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	File::File()
		: hdl(INVALID_HANDLE_VALUE), map(INVALID_HANDLE_VALUE), view(nullptr), viewSize(0)
	{
	}

	File::File(const Str_8 &filePath, const Mode mode, const Disposition disposition)
		: BaseFile(filePath, mode, disposition), hdl(INVALID_HANDLE_VALUE), map(INVALID_HANDLE_VALUE), view(nullptr),
		viewSize(0)
	{
		DWORD winMode = 0;
		switch (mode)
		{
			case Mode::READ:
				winMode = GENERIC_READ;
				break;
			case Mode::WRITE:
				winMode = GENERIC_WRITE;
				break;
			case Mode::READ_WRITE:
				winMode = GENERIC_READ | GENERIC_WRITE;
				break;
		}

		DWORD winDisp = 0;
		switch (disposition)
		{
			case Disposition::CREATE_PERSISTENT:
				winDisp = CREATE_ALWAYS;
				break;
			case Disposition::CREATE:
				winDisp = CREATE_NEW;
				break;
			case Disposition::OPEN_PERSISTENT:
				winDisp = OPEN_ALWAYS;
				break;
			case Disposition::OPEN:
				winDisp = OPEN_EXISTING;
				break;
			case Disposition::TRUNCATE:
				winDisp = TRUNCATE_EXISTING;
				break;
		}

		hdl = CreateFileW(UTF::To_16(path), winMode, 0, nullptr, winDisp, FILE_ATTRIBUTE_NORMAL, nullptr);
		if (hdl == INVALID_HANDLE_VALUE)
		{
			DWORD code = GetLastError();
			if (code == ERROR_FILE_NOT_FOUND)
				EHS_LOG_INT(LogType::ERR, 0, "File not found at path, \"" + path + "\".");
			else if (code != ERROR_SUCCESS)
				EHS_LOG_INT(LogType::ERR, 1, "Failed to create handle for file, \"" + path + "\", with error #" + Str_8::FromNum(code) + ".");

			return;
		}

		EHS_LOG_SUCCESS();
	}

	File::File(File&& file) noexcept
		: BaseFile(std::move(file)), hdl(file.hdl), map(file.map), view(file.view), viewSize(file.viewSize)
	{
		file.hdl = INVALID_HANDLE_VALUE;
		file.map = INVALID_HANDLE_VALUE;
		file.view = nullptr;
		file.viewSize = 0;
	}

	File::File(const File& file)
		: BaseFile(file), hdl(INVALID_HANDLE_VALUE), map(INVALID_HANDLE_VALUE), view(nullptr), viewSize(0)
	{
	}

	File& File::operator=(File&& file) noexcept
	{
		if (this == &file)
			return *this;

		BaseFile::operator=(std::move(file));

		hdl = file.hdl;
		map = file.map;
		view = file.view;
		viewSize = file.viewSize;

		file.hdl = INVALID_HANDLE_VALUE;
		file.map = INVALID_HANDLE_VALUE;
		file.view = nullptr;
		file.viewSize = 0;

		return *this;
	}

	File& File::operator=(const File& file)
	{
		if (this == &file)
			return *this;

		BaseFile::operator=(file);

		hdl = INVALID_HANDLE_VALUE;
		map = INVALID_HANDLE_VALUE;
		view = nullptr;
		viewSize = 0;

		return *this;
	}

	File::operator const Byte*() const
	{
		return view;
	}

	File::operator Byte*()
	{
		return view;
	}

	void File::Release()
	{
		if (view && !UnmapViewOfFile(view))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to unmap view with error #" + Str_8::FromNum(GetLastError()) + ".");
		view = nullptr;
		viewSize = 0;

		if (IsMapped() && !CloseHandle(map))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to unmap with error #" + Str_8::FromNum(GetLastError()) + ".");
		map = INVALID_HANDLE_VALUE;

		if (IsValid() && !CloseHandle(hdl))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to close file handle with error #" + Str_8::FromNum(GetLastError()) + ".");
		hdl = INVALID_HANDLE_VALUE;
	}

	bool File::IsMapped() const
	{
		return map != INVALID_HANDLE_VALUE && view;
	}

	UInt_64 File::MapSize() const
	{
		return viewSize;
	}

	void File::Map(const UInt_64 offset, const UInt_64 size)
	{
		if (!IsValid() || IsMapped())
			return;

		DWORD winMode = 0;
		switch (mode)
		{
			case Mode::READ:
				winMode = PAGE_READONLY;
				break;
			case Mode::WRITE:
				winMode = PAGE_READWRITE;
				break;
			case Mode::READ_WRITE:
				winMode = PAGE_READWRITE;
				break;
		}

		map = CreateFileMappingW(hdl, nullptr, winMode, 0, 0, nullptr);
		if (!map)
		{
			EHS_LOG_INT(LogType::ERR, 0, "Failed to create map handle with error #" + Str_8::FromNum(GetLastError()) + ".");
			return;
		}

		switch (mode)
		{
			case Mode::READ:
				winMode = FILE_MAP_READ;
				break;
			case Mode::WRITE:
				winMode = FILE_MAP_WRITE;
				break;
			case Mode::READ_WRITE:
				winMode = FILE_MAP_ALL_ACCESS;
				break;
		}

		view = (Byte*)MapViewOfFile(map, winMode, ((DWORD*)&offset)[1], (DWORD)offset, size);
		if (!view)
		{
			EHS_LOG_INT(LogType::ERR, 0, "Failed to map view with error #" + Str_8::FromNum(GetLastError()) + ".");

			if (!CloseHandle(map))
				EHS_LOG_INT(LogType::ERR, 0, "Failed to unmap with error #" + Str_8::FromNum(GetLastError()) + ".");
			map = INVALID_HANDLE_VALUE;
		}

		viewSize = size;
	}

	void File::Unmap()
	{
		if (!IsValid() || !IsMapped())
			return;

		if (!UnmapViewOfFile(view))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to unmap view with error #" + Str_8::FromNum(GetLastError()) + ".");
		view = nullptr;
		viewSize = 0;

		if (!CloseHandle(map))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to unmap with error #" + Str_8::FromNum(GetLastError()) + ".");
		map = INVALID_HANDLE_VALUE;
	}

	void File::FlushMap()
	{
		if (!IsValid() || !IsMapped())
			return;

		if (!FlushViewOfFile(view, viewSize))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to flush view with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	UInt_64 File::Write(const Byte *const data, const UInt_64 size)
	{
		if (!IsValid() || IsMapped())
			return 0;

		SInt_64 written = 0;

		if (!WriteFile(hdl, data, (DWORD)size, (DWORD*)&written, nullptr))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to write to file, \"" + path + "\", with error #" + Str_8::FromNum(GetLastError()) + ".");

		return (UInt_64)written;
	}

	UInt_64 File::Read(Byte *const data, const UInt_64 size)
	{
		if (!IsValid() || IsMapped())
			return 0;

		SInt_64 read = 0;

		if (!ReadFile(hdl, data, (DWORD)size, (DWORD*)&read, nullptr))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to read from file, \"" + path + "\", with error #" + Str_8::FromNum(GetLastError()) + ".");

		return (UInt_64)read;
	}

	void File::Seek(UInt_64 index)
	{
		if (!IsValid() || IsMapped())
			return;

		if (SetFilePointer(hdl, (LONG)index, (PLONG)&((UInt_32*)&index)[1], FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			EHS_LOG_INT(LogType::ERR, 0, "Failed to seek with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	void File::SeekBeginning()
	{
		if (!IsValid() || IsMapped())
			return;

		if (SetFilePointer(hdl, 0, nullptr, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			EHS_LOG_INT(LogType::ERR, 0, "Failed to seek with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	void File::SeekEnd()
	{
		if (!IsValid() || IsMapped())
			return;

		if (SetFilePointer(hdl, 0, nullptr, FILE_END) == INVALID_SET_FILE_POINTER)
			EHS_LOG_INT(LogType::ERR, 0, "Failed to seek with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	void File::Truncate(const UInt_64 size)
	{
		if (!IsValid() || IsMapped())
			return;

		Seek(size);

		if (!::SetEndOfFile(hdl))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to set end of file with error #" + Str_8::FromNum(GetLastError()) + ".");

		SeekBeginning();
	}

	UInt_64 File::Size() const
	{
		if (!IsValid())
			return 0;

		LARGE_INTEGER size = {};

		if (!GetFileSizeEx(hdl, &size))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to retrieve file size with error #" + Str_8::FromNum(GetLastError()) + ".");

		return (UInt_64)size.QuadPart;
	}

	bool File::IsValid() const
	{
		return hdl != INVALID_HANDLE_VALUE;
	}

	void File::Rename_32(const Str_32& filePath, const Str_32& newName)
	{
		Rename_16(UTF::To_16(filePath), UTF::To_16(newName));
	}

	void File::Rename_16(const Str_16& filePath, const Str_16& newName)
	{
		UInt_64 index = 0;
		Str_16 path;

		if (filePath.Find(L"/", &index, SearchPattern::RIGHT_LEFT) || filePath.Find(L"\\", &index, SearchPattern::RIGHT_LEFT))
			path = filePath.Sub(0, index);

		if (!MoveFileW(filePath, path + newName))
			EHS_LOG_INT(LogType::ERR, 0, "Failed to rename file with error #" + Str_8::FromNum(GetLastError()) + ".");
	}

	void File::Rename_8(const Str_8& filePath, const Str_8& newName)
	{
		Rename_16(UTF::To_16(filePath), UTF::To_16(newName));
	}
}
